# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A thread pool that logs exceptions raised by tasks executed within it."""

from concurrent import futures
import logging

_LOGGER = logging.getLogger(__name__)


def _wrap(behavior):
    """Wraps an arbitrary callable behavior in exception-logging."""

    def _wrapping(*args, **kwargs):
        try:
            return behavior(*args, **kwargs)
        except Exception:
            _LOGGER.exception(
                "Unexpected exception from %s executed in logging pool!",
                behavior,
            )
            raise

    return _wrapping


class _LoggingPool(object):
    """An exception-logging futures.ThreadPoolExecutor-compatible thread pool."""

    def __init__(self, backing_pool):
        self._backing_pool = backing_pool

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._backing_pool.shutdown(wait=True)

    def submit(self, fn, *args, **kwargs):
        return self._backing_pool.submit(_wrap(fn), *args, **kwargs)

    def map(self, func, *iterables, **kwargs):
        return self._backing_pool.map(
            _wrap(func), *iterables, timeout=kwargs.get("timeout", None)
        )

    def shutdown(self, wait=True):
        self._backing_pool.shutdown(wait=wait)


def pool(max_workers):
    """Creates a thread pool that logs exceptions raised by the tasks within it.

    Args:
      max_workers: The maximum number of worker threads to allow the pool.

    Returns:
      A futures.ThreadPoolExecutor-compatible thread pool that logs exceptions
        raised by the tasks executed within it.
    """
    return _LoggingPool(futures.ThreadPoolExecutor(max_workers))



from messageprocessorutil import UuidMessageProcessor
from messagebus import NakMessage, AckMessage


class OrderModifiedIgnoredTypesProcessor(UuidMessageProcessor):
    def __init__(self, ):
        # type: () -> None

        super(OrderModifiedIgnoredTypesProcessor, self).__init__()

    def parse_data(self, data, event_type):
        return None

    def call_business(self, _):
        return None

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_event(event, AckMessage())

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_event(event, NakMessage())

    def format_parse_exception(self, message_bus, message, event, data, exception):
        self.format_exception(message_bus, message, event, data, exception)



#!jython
#
# Backend Jython with JavaComm
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2002-2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

from __future__ import absolute_import

from serial.serialutil import *


def my_import(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def detect_java_comm(names):
    """try given list of modules and return that imports"""
    for name in names:
        try:
            mod = my_import(name)
            mod.SerialPort
            return mod
        except (ImportError, AttributeError):
            pass
    raise ImportError("No Java Communications API implementation found")


# Java Communications API implementations
# http://mho.republika.pl/java/comm/

comm = detect_java_comm([
    'javax.comm',  # Sun/IBM
    'gnu.io',      # RXTX
])


def device(portnumber):
    """Turn a port number into a device name"""
    enum = comm.CommPortIdentifier.getPortIdentifiers()
    ports = []
    while enum.hasMoreElements():
        el = enum.nextElement()
        if el.getPortType() == comm.CommPortIdentifier.PORT_SERIAL:
            ports.append(el)
    return ports[portnumber].getName()


class Serial(SerialBase):
    """\
    Serial port class, implemented with Java Communications API and
    thus usable with jython and the appropriate java extension.
    """

    def open(self):
        """\
        Open port with current settings. This may throw a SerialException
        if the port cannot be opened.
        """
        if self._port is None:
            raise SerialException("Port must be configured before it can be used.")
        if self.is_open:
            raise SerialException("Port is already open.")
        if type(self._port) == type(''):      # strings are taken directly
            portId = comm.CommPortIdentifier.getPortIdentifier(self._port)
        else:
            portId = comm.CommPortIdentifier.getPortIdentifier(device(self._port))     # numbers are transformed to a comport id obj
        try:
            self.sPort = portId.open("python serial module", 10)
        except Exception as msg:
            self.sPort = None
            raise SerialException("Could not open port: %s" % msg)
        self._reconfigurePort()
        self._instream = self.sPort.getInputStream()
        self._outstream = self.sPort.getOutputStream()
        self.is_open = True

    def _reconfigurePort(self):
        """Set communication parameters on opened port."""
        if not self.sPort:
            raise SerialException("Can only operate on a valid port handle")

        self.sPort.enableReceiveTimeout(30)
        if self._bytesize == FIVEBITS:
            jdatabits = comm.SerialPort.DATABITS_5
        elif self._bytesize == SIXBITS:
            jdatabits = comm.SerialPort.DATABITS_6
        elif self._bytesize == SEVENBITS:
            jdatabits = comm.SerialPort.DATABITS_7
        elif self._bytesize == EIGHTBITS:
            jdatabits = comm.SerialPort.DATABITS_8
        else:
            raise ValueError("unsupported bytesize: %r" % self._bytesize)

        if self._stopbits == STOPBITS_ONE:
            jstopbits = comm.SerialPort.STOPBITS_1
        elif self._stopbits == STOPBITS_ONE_POINT_FIVE:
            jstopbits = comm.SerialPort.STOPBITS_1_5
        elif self._stopbits == STOPBITS_TWO:
            jstopbits = comm.SerialPort.STOPBITS_2
        else:
            raise ValueError("unsupported number of stopbits: %r" % self._stopbits)

        if self._parity == PARITY_NONE:
            jparity = comm.SerialPort.PARITY_NONE
        elif self._parity == PARITY_EVEN:
            jparity = comm.SerialPort.PARITY_EVEN
        elif self._parity == PARITY_ODD:
            jparity = comm.SerialPort.PARITY_ODD
        elif self._parity == PARITY_MARK:
            jparity = comm.SerialPort.PARITY_MARK
        elif self._parity == PARITY_SPACE:
            jparity = comm.SerialPort.PARITY_SPACE
        else:
            raise ValueError("unsupported parity type: %r" % self._parity)

        jflowin = jflowout = 0
        if self._rtscts:
            jflowin |= comm.SerialPort.FLOWCONTROL_RTSCTS_IN
            jflowout |= comm.SerialPort.FLOWCONTROL_RTSCTS_OUT
        if self._xonxoff:
            jflowin |= comm.SerialPort.FLOWCONTROL_XONXOFF_IN
            jflowout |= comm.SerialPort.FLOWCONTROL_XONXOFF_OUT

        self.sPort.setSerialPortParams(self._baudrate, jdatabits, jstopbits, jparity)
        self.sPort.setFlowControlMode(jflowin | jflowout)

        if self._timeout >= 0:
            self.sPort.enableReceiveTimeout(int(self._timeout*1000))
        else:
            self.sPort.disableReceiveTimeout()

    def close(self):
        """Close port"""
        if self.is_open:
            if self.sPort:
                self._instream.close()
                self._outstream.close()
                self.sPort.close()
                self.sPort = None
            self.is_open = False

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def in_waiting(self):
        """Return the number of characters currently in the input buffer."""
        if not self.sPort:
            raise PortNotOpenError()
        return self._instream.available()

    def read(self, size=1):
        """\
        Read size bytes from the serial port. If a timeout is set it may
        return less characters as requested. With no timeout it will block
        until the requested number of bytes is read.
        """
        if not self.sPort:
            raise PortNotOpenError()
        read = bytearray()
        if size > 0:
            while len(read) < size:
                x = self._instream.read()
                if x == -1:
                    if self.timeout >= 0:
                        break
                else:
                    read.append(x)
        return bytes(read)

    def write(self, data):
        """Output the given string over the serial port."""
        if not self.sPort:
            raise PortNotOpenError()
        if not isinstance(data, (bytes, bytearray)):
            raise TypeError('expected %s or bytearray, got %s' % (bytes, type(data)))
        self._outstream.write(data)
        return len(data)

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self.sPort:
            raise PortNotOpenError()
        self._instream.skip(self._instream.available())

    def reset_output_buffer(self):
        """\
        Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        if not self.sPort:
            raise PortNotOpenError()
        self._outstream.flush()

    def send_break(self, duration=0.25):
        """Send break condition. Timed, returns to idle state after given duration."""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.sendBreak(duration*1000.0)

    def _update_break_state(self):
        """Set break: Controls TXD. When active, to transmitting is possible."""
        if self.fd is None:
            raise PortNotOpenError()
        raise SerialException("The _update_break_state function is not implemented in java.")

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.setRTS(self._rts_state)

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.setDTR(self._dtr_state)

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.isCTS()

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.isDSR()

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.isRI()

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        if not self.sPort:
            raise PortNotOpenError()
        self.sPort.isCD()



from typing import List
from ..model import OrderData
from old_helper import BaseRepository
from msgbus import MBEasyContext
from persistence import Connection
import base64


class OrderRetrieverRepository(BaseRepository):
    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        super(OrderRetrieverRepository, self).__init__(mbcontext)

    def get_order_data_list(self, initial_order_id, last_order_id):
        # type: (int, int) -> List[OrderData]
        def inner_func(conn):
            # type: (Connection) -> List[OrderData]
            if last_order_id is not None:
                query = self._RetrieveOrderQueryBetween.format(initial_order_id, last_order_id)
            else:
                query = self._RetrieveOrderQueryHigher.format(initial_order_id)

            ret = []
            for x in conn.select(query):
                try:
                    order_data = OrderData(int(x.get_entry(0)),
                                           int(x.get_entry(1)),
                                           base64.b64decode(x.get_entry(2)),
                                           base64.b64decode(x.get_entry(3)))
                    ret.append(order_data)
                except Exception:
                    pass

            return ret

        return self.execute_with_connection(inner_func, service_name="FiscalPersistence")

    def get_last_order_id(self):
        # type: () -> int
        def inner_func(conn):
            # type: (Connection) -> int
            cursor = conn.select(self._LastOrderQuery)
            if cursor.rows() > 0:
                row_data = cursor.get_row(0).get_entry(0)
                if row_data is None:
                    return 0
                else:
                    return int(row_data)
            else:
                return 0

        return self.execute_with_connection(inner_func, service_name="FiscalPersistence")

    _RetrieveOrderQueryBetween = """SELECT OrderId, PosId, OrderPicture, XmlRequest 
FROM FiscalData
WHERE (SentToNfce = 1) AND (OrderId BETWEEN '{0}' AND '{1}')
ORDER BY OrderId"""

    _RetrieveOrderQueryHigher = """SELECT OrderId, PosId, OrderPicture, XmlRequest
FROM FiscalData
WHERE (SentToNfce = 1) AND (OrderId > '{0}')
ORDER BY OrderId"""

    _LastOrderQuery = """SELECT max(OrderId) FROM FiscalData"""



# aadhaar.py - functions for handling Indian personal identity numbers
#
# Copyright (C) 2017 Srikanth L
# Copyright (C) 2021 Gaurav Chauhan
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Aadhaar (Indian personal identity number).

Aadhaar is a 12 digit identification number that can be obtained by Indian
citizens, non-residents passport holders of India and resident foreign
nationals. The number is issued by the Unique Identification Authority of
India (UIDAI).

Aadhaar is made up of 12 digits where the last digits is a check digit
calculated using the Verhoeff algorithm. The numbers are generated in a
random, non-repeating sequence and do not begin with 0 or 1.

More information:

* https://en.wikipedia.org/wiki/Aadhaar
* https://web.archive.org/web/20140611025606/http://uidai.gov.in/UID_PDF/Working_Papers/A_UID_Numbering_Scheme.pdf

>>> validate('234123412346')
'234123412346'
>>> validate('234123412347')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('123412341234')  # number should not start with 0 or 1
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('643343121')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('222222222222')  # number cannot be a palindrome
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('234123412346')
'2341 2341 2346'
>>> mask('234123412346')
'XXXX XXXX 2346'
"""

import re

from stdnum import verhoeff
from stdnum.exceptions import *
from stdnum.util import clean


aadhaar_re = re.compile(r'^[2-9][0-9]{11}$')
"""Regular expression used to check syntax of Aadhaar numbers."""


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def validate(number):
    """Check if the number provided is a valid Aadhaar number. This checks
    the length, formatting and check digit."""
    number = compact(number)
    if len(number) != 12:
        raise InvalidLength()
    if not aadhaar_re.match(number):
        raise InvalidFormat()
    if number == number[::-1]:
        raise InvalidFormat()  # Aadhaar cannot be a palindrome
    verhoeff.validate(number)
    return number


def is_valid(number):
    """Check if the number provided is a valid Aadhaar number. This checks
    the length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:4], number[4:8], number[8:]))


def mask(number):
    """Masks the first 8 digits as per Ministry of Electronics and
    Information Technology (MeitY) guidelines."""
    number = compact(number)
    return 'XXXX XXXX ' + number[-4:]



"""
SignXML exception types.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import cryptography.exceptions

class InvalidSignature(cryptography.exceptions.InvalidSignature):
    """
    Raised when signature validation fails.
    """

class InvalidDigest(InvalidSignature):
    """
    Raised when digest validation fails (causing the signature to be untrusted).
    """

class InvalidCertificate(InvalidSignature):
    """
    Raised when certificate validation fails.
    """

class InvalidInput(ValueError):
    pass

class RedundantCert(Exception):
    pass



# -*- coding: utf-8 -*-
import json
import logging

from remove_coupon import remove_discounts
from mw_helper import (
    show_message_dialog,
    show_check_list_box_dialog
)
from sysactions import (
    get_model,
    get_current_order,
    format_amount,
    show_messagebox,
    get_posot,
    translate_message,
    StopAction
)
from xml.etree import cElementTree as eTree

from typing import (
    Dict,
    List,
    Union,
)

logger = logging.getLogger("PosActions")


def display_coupons(pos_id, table_id=None):
    model = get_model(pos_id)
    pos_ot = get_posot(model)
    order = get_current_order(model)
    if order is None:
        return

    try:
        benefits = get_benefits_from_order(pos_ot, order)
        if benefits is None or len(benefits) == 0:
            show_messagebox(pos_id, "$ORDER_HAS_NO_APPLIED_DISCOUNT")
            return

        applied_coupons = _get_applied_coupons_from_benefits(
            model=model,
            benefits=benefits,
        )
        selected_indexes = _select_coupon(pos_id, applied_coupons)
        if not len(selected_indexes):
            return

        coupon_ids = []
        for selected_index in selected_indexes:
            index = int(selected_index)
            if index == 0:
                continue

            coupon = [x for x in applied_coupons if x["index"] == index]
            if not coupon:
                continue

            disable = coupon[0]["name"].split("##")[2]
            if not eval(disable):
                coupon_ids.append(coupon[0]["id"])

        remove_discounts(
            pos_id,
            coupon_ids
        )
        if table_id is not None:
            from table_actions import process_discount_in_table, clean_table_payments, create_table_modified_event

            table_picture = clean_table_payments(pos_id, table_id, delete_order_tender=False)
            create_table_modified_event(pos_id, table_picture)

            process_discount_in_table(pos_id, order, table_id)

        return True

    except StopAction:
        raise
    except Exception as ex:
        show_message_dialog(pos_id, "$ERROR", ex.message)
        logger.exception("Error removing discount")
        return False


def get_benefits_from_order(pos_ot, order, custom_properties=None):
    if not order:
        return []

    if not custom_properties:
        custom_properties = eTree.XML(pos_ot.getOrderCustomProperties(orderid=order.get("orderId")))

    benefit_list_custom_property = custom_properties.find(".//OrderProperty[@key='BENEFIT_LIST']")
    if benefit_list_custom_property is None:
        return []

    benefit_list_json = json.loads(benefit_list_custom_property.get("value"))
    benefit_json = benefit_list_json.get("benefitList")
    benefit_list = json.loads(benefit_json)
    benefit_list = [b for b in benefit_list if b["benefit_applied"] or b["added_sale_lines"]]

    return benefit_list


def _get_applied_coupons_from_benefits(
    model,      # type: eTree.Element
    benefits,   # type: List[Dict[str, Union[int, bool, float, str, Dict]]]
):
    # type: (...) -> List[Dict[str, Union[int, str, None]]]

    all_coupons = translate_message(model, "ALL_COUPONS")
    applied_coupons = []
    total_amount = 0
    benefit_index = 1

    for benefit in benefits:
        benefit_applied = benefit.get("benefit_applied", False)
        if not benefit_applied:
            continue

        benefit_detail = benefit["benefit"]
        benefit_total_value = benefit["benefit_total_value"]
        if benefit_detail:
            added_sale_lines = len(benefit.get("added_sale_lines", []))
            if added_sale_lines and benefit_total_value <= 0:
                continue

            benefit_key = benefit["benefit_key"]
            benefit_description = benefit_detail.get("userDescription", benefit_detail.get("id", ""))
            description = benefit_description.encode("UTF-8")
            benefit_added_in_progress = benefit["order_state"] == "IN_PROGRESS"
            disable = benefit_added_in_progress or added_sale_lines > 0
            formatted_amount = format_amount(
                model=model,
                amount=benefit_total_value,
                addsymbol=False,
            )
            name = "{}##{}##{}".format(description, formatted_amount, disable)
            data = {"index": benefit_index, "id": benefit_key, "name": name}
            applied_coupons.append(data)

        benefit_index += 1
        total_amount += benefit_total_value

    name = "{}##{}##False".format(all_coupons, format_amount(model, total_amount, False))
    data = {"index": 0, "id": None, "name": name}
    applied_coupons.append(data)
    applied_coupons.sort(key=lambda x: (x["index"]))
    return applied_coupons


def _select_coupon(pos_id, coupons):
    options = ["{0}".format(c["name"]) for c in coupons]
    return show_check_list_box_dialog(pos_id, options, "$SELECT_ONE_COUPON", buttons="$REMOVE|$PREVIOUS")



# associationid.py - functions for handling Finnish association registry id
# coding: utf-8
#
# Copyright (C) 2015 Holvi Payment Services Oy
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Finnish Association Identifier.

The number consists of 1 to 6 digits that are normally separated with a dot
in groups of 0-3 and 0-3 numbers. E.g. 123.123, 12.123, 1.123, 123 or 1.

>>> validate('123.123')
'123123'
>>> validate('1123')
'1123'
>>> validate('123123123')
Traceback (most recent call last):
  ...
InvalidLength: The number has an invalid length.
>>> validate('12df')
Traceback (most recent call last):
  ...
InvalidFormat: The number has an invalid format.
>>> format('123')
'123'
>>> format('1234')
'1.234'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# a collection of all registered numbers with 2 or less digits
_lownumbers = set((
    1, 6, 7, 9, 12, 14, 15, 16, 18, 22, 23, 24, 27, 28, 29, 35, 36, 38, 40,
    41, 42, 43, 45, 46, 50, 52, 55, 58, 60, 64, 65, 68, 72, 75, 76, 77, 78,
    83, 84, 85, 89, 92))


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -._+').strip()


def validate(number):
    """Check if the number is a valid Finnish association register number.
    This checks the length and format."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) < 1 or len(number) > 6:
        raise InvalidLength()
    if len(number) < 3 and int(number) not in _lownumbers:
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid association register number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    if len(number) <= 3:
        return number
    else:
        return number[:-3] + '.' + number[-3:]



import time
from logging import Logger

from threading import Lock, Thread, local
from messageprocessor import MessageProcessorCallback


class TimeProcessorCallback(MessageProcessorCallback):
    def __init__(self, logger, log_time):
        # type: (Logger, float) -> None
        self.logger = logger
        self.log_time = log_time

        self.count_lock = Lock()
        self.current_count = 0
        self.process_time = 0
        self.local_storage = local()

        t = Thread(target=self.print_result)
        t.daemon = True
        t.start()

    def process_begun(self, event_name, request_id, **kwargs):
        self.local_storage.start_time = time.process_time()

    def input_parsed(self, event_name, request_id, input_obj):
        return

    def business_called(self, event_name, request_id, input_obj, output_obj):
        return

    def process_finished(self, event_name, request_id, input_obj, output_obj):
        self.calculate_end()

    def parse_exception(self, event_name, request_id, data, exception):
        self.calculate_end()

    def unhandled_exception(self, event_name, request_id, input_obj, exception):
        self.calculate_end()

    def calculate_end(self):
        end = time.process_time()
        with self.count_lock:
            self.process_time += end - self.local_storage.start_time
            self.current_count += 1

    def print_result(self):
        while True:
            with self.count_lock:
                process_time = self.process_time
                current_count = self.current_count
                self.process_time = 0
                self.current_count = 0

            if current_count > 0:
                self.logger.info("Count: {}, Process average time: {} ms".format(
                    current_count,
                    int((process_time / current_count) * 1000)))

            time.sleep(self.log_time)



# amka.py - functions for handling Greek social security numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""AMKA (Αριθμός Μητρώου Κοινωνικής Ασφάλισης, Greek social security number).

The Αριθμός Μητρώου Κοινωνικής Ασφάλισης (AMKA or Arithmos Mitroou Koinonikis
Asfalisis) is the personal identifier that is used for social security
purposes in Greece. The number consists of 11 digits and includes the
person's date of birth and gender.

More information:

* https://www.amka.gr/tieinai_en.html

>>> validate('01013099997')
'01013099997'
>>> validate('01013099999')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> get_birth_date('01013099997')
datetime.date(1930, 1, 1)
>>> get_gender('01013099997')
'M'
"""

import datetime

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def get_birth_date(number):
    """Split the date parts from the number and return the date of birth.
    Since only two digits are used for the year, the century may be
    incorrect."""
    number = compact(number)
    day = int(number[0:2])
    month = int(number[2:4])
    year = int(number[4:6]) + 1900
    try:
        return datetime.date(year, month, day)
    except ValueError:
        try:
            return datetime.date(year + 100, month, day)
        except ValueError:
            raise InvalidComponent()


def get_gender(number):
    """Get the gender (M/F) from the person's AMKA."""
    number = compact(number)
    if int(number[9]) % 2:
        return 'M'
    else:
        return 'F'


def validate(number):
    """Check if the number is a valid AMKA. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    luhn.validate(number)
    get_birth_date(number)
    return number


def is_valid(number):
    """Check if the number is a valid AMKA."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# rif.py - functions for handling Venezuelan VAT numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RIF (Registro de Identificación Fiscal, Venezuelan VAT number).

The Registro de Identificación Fiscal (RIF) is the Venezuelan fiscal
registration number. The number consists of 10 digits where the first digit
denotes the type of number (person, company or government) and the last digit
is a check digit.

>>> validate('V-11470283-4')
'V114702834'
>>> validate('V-11470283-3')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


# Known number types and their corresponding value in the check
# digit calculation
_company_types = {
    'V': 4,   # natural person born in Venezuela
    'E': 8,   # foreign natural person
    'J': 12,  # company
    'P': 16,  # passport
    'G': 20,  # government
}


def calc_check_digit(number):
    """Calculate the check digit for the RIF."""
    number = compact(number)
    weights = (3, 2, 7, 6, 5, 4, 3, 2)
    c = _company_types[number[0]]
    c += sum(w * int(n) for w, n in zip(weights, number[1:9]))
    return '00987654321'[c % 11]


def validate(number):
    """Check if the number provided is a valid RIF. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 10:
        raise InvalidLength()
    if number[0] not in _company_types:
        raise InvalidComponent()
    if not isdigits(number[1:]):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid RIF. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# hp.py - functions for handling Israeli company numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Company Number (מספר חברה, or short ח.פ. Israeli company number).

It consists of nine digits and includes a check digit. For companies
the first digit is a 5. The first two digits identify the type of
company.

More information:

* https://he.wikipedia.org/wiki/תאגיד#מספר_רישום_התאגיד
* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Israel-TIN.pdf
* https://wiki.scn.sap.com/wiki/display/CRM/Israel

>>> validate('516179157')
'516179157'
>>> format(' 5161 79157 ')
'516179157'
>>> validate('516179150')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('490154203237518')  # longer than 9 digits
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('416179157')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""  # noqa: E501

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def validate(number):
    """Check if the number provided is a valid ID. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not isdigits(number) or int(number) <= 0:
        raise InvalidFormat()
    if number[0] != '5':
        raise InvalidComponent()
    luhn.validate(number)
    return number


def is_valid(number):
    """Check if the number provided is a valid ID. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



#!/usr/bin/env python
"""
qr - Convert stdin (or the first argument) to a QR Code.

When stdout is a tty the QR Code is printed to the terminal and when stdout is
a pipe to a file an image is written. The default image format is PNG.
"""
import sys
import optparse
import os
import qrcode
# The next block is added to get the terminal to display properly on MS platforms
if sys.platform.startswith(('win', 'cygwin')):
    import colorama
    colorama.init()

default_factories = {
    'pil': 'qrcode.image.pil.PilImage',
    'pymaging': 'qrcode.image.pure.PymagingImage',
    'svg': 'qrcode.image.svg.SvgImage',
    'svg-fragment': 'qrcode.image.svg.SvgFragmentImage',
    'svg-path': 'qrcode.image.svg.SvgPathImage',
}

error_correction = {
    'L': qrcode.ERROR_CORRECT_L,
    'M': qrcode.ERROR_CORRECT_M,
    'Q': qrcode.ERROR_CORRECT_Q,
    'H': qrcode.ERROR_CORRECT_H,
}


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    from pkg_resources import get_distribution
    version = get_distribution('qrcode').version
    parser = optparse.OptionParser(usage=__doc__.strip(), version=version)
    parser.add_option(
        "--factory", help="Full python path to the image factory class to "
        "create the image with. You can use the following shortcuts to the "
        "built-in image factory classes: {0}.".format(
            ", ".join(sorted(default_factories.keys()))))
    parser.add_option(
        "--optimize", type=int, help="Optimize the data by looking for chunks "
        "of at least this many characters that could use a more efficient "
        "encoding method. Use 0 to turn off chunk optimization.")
    parser.add_option(
        "--error-correction", type='choice',
        choices=sorted(error_correction.keys()), default='M',
        help="The error correction level to use. Choices are L (7%), "
        "M (15%, default), Q (25%), and H (30%).")
    opts, args = parser.parse_args(args)

    qr = qrcode.QRCode(
        error_correction=error_correction[opts.error_correction])

    if opts.factory:
        module = default_factories.get(opts.factory, opts.factory)
        if '.' not in module:
            parser.error("The image factory is not a full python path")
        module, name = module.rsplit('.', 1)
        imp = __import__(module, {}, [], [name])
        image_factory = getattr(imp, name)
    else:
        image_factory = None

    if args:
        data = args[0]
    else:
        # Use sys.stdin.buffer if available (Python 3) avoiding
        # UnicodeDecodeErrors.
        stdin_buffer = getattr(sys.stdin, 'buffer', sys.stdin)
        data = stdin_buffer.read()
    if opts.optimize is None:
        qr.add_data(data)
    else:
        qr.add_data(data, optimize=opts.optimize)

    if image_factory is None and os.isatty(sys.stdout.fileno()):
        qr.print_ascii(tty=True)
        return

    img = qr.make_image(image_factory=image_factory)

    sys.stdout.flush()
    # Use sys.stdout.buffer if available (Python 3), avoiding
    # UnicodeDecodeErrors.
    stdout_buffer = getattr(sys.stdout, 'buffer', None)
    if not stdout_buffer:
        if sys.platform == 'win32':  # pragma: no cover
            import msvcrt
            msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
        stdout_buffer = sys.stdout

    img.save(stdout_buffer)


if __name__ == "__main__":
    main()



from __future__ import annotations

import contextlib
import re
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from sqlalchemy import __version__
from sqlalchemy import inspect
from sqlalchemy import schema
from sqlalchemy import sql
from sqlalchemy import types as sqltypes
from sqlalchemy.engine import url
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import CheckConstraint
from sqlalchemy.schema import Column
from sqlalchemy.schema import ForeignKeyConstraint
from sqlalchemy.sql import visitors
from sqlalchemy.sql.base import DialectKWArgs
from sqlalchemy.sql.elements import BindParameter
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.elements import quoted_name
from sqlalchemy.sql.elements import TextClause
from sqlalchemy.sql.elements import UnaryExpression
from sqlalchemy.sql.visitors import traverse
from typing_extensions import TypeGuard

if TYPE_CHECKING:
    from sqlalchemy import Index
    from sqlalchemy import Table
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine import Dialect
    from sqlalchemy.engine import Transaction
    from sqlalchemy.engine.reflection import Inspector
    from sqlalchemy.sql.base import ColumnCollection
    from sqlalchemy.sql.compiler import SQLCompiler
    from sqlalchemy.sql.dml import Insert
    from sqlalchemy.sql.elements import ColumnElement
    from sqlalchemy.sql.schema import Constraint
    from sqlalchemy.sql.schema import SchemaItem
    from sqlalchemy.sql.selectable import Select
    from sqlalchemy.sql.selectable import TableClause

_CE = TypeVar("_CE", bound=Union["ColumnElement[Any]", "SchemaItem"])


def _safe_int(value: str) -> Union[int, str]:
    try:
        return int(value)
    except:
        return value


_vers = tuple(
    [_safe_int(x) for x in re.findall(r"(\d+|[abc]\d)", __version__)]
)
sqla_13 = _vers >= (1, 3)
sqla_14 = _vers >= (1, 4)
# https://docs.sqlalchemy.org/en/latest/changelog/changelog_14.html#change-0c6e0cc67dfe6fac5164720e57ef307d
sqla_14_18 = _vers >= (1, 4, 18)
sqla_14_26 = _vers >= (1, 4, 26)
sqla_2 = _vers >= (2,)
sqlalchemy_version = __version__

try:
    from sqlalchemy.sql.naming import _NONE_NAME as _NONE_NAME
except ImportError:
    from sqlalchemy.sql.elements import _NONE_NAME as _NONE_NAME  # type: ignore  # noqa: E501


class _Unsupported:
    "Placeholder for unsupported SQLAlchemy classes"


try:
    from sqlalchemy import Computed
except ImportError:
    if not TYPE_CHECKING:

        class Computed(_Unsupported):
            pass

    has_computed = False
    has_computed_reflection = False
else:
    has_computed = True
    has_computed_reflection = _vers >= (1, 3, 16)

try:
    from sqlalchemy import Identity
except ImportError:
    if not TYPE_CHECKING:

        class Identity(_Unsupported):
            pass

    has_identity = False
else:
    identity_has_dialect_kwargs = issubclass(Identity, DialectKWArgs)

    def _get_identity_options_dict(
        identity: Union[Identity, schema.Sequence, None],
        dialect_kwargs: bool = False,
    ) -> Dict[str, Any]:
        if identity is None:
            return {}
        elif identity_has_dialect_kwargs:
            as_dict = identity._as_dict()  # type: ignore
            if dialect_kwargs:
                assert isinstance(identity, DialectKWArgs)
                as_dict.update(identity.dialect_kwargs)
        else:
            as_dict = {}
            if isinstance(identity, Identity):
                # always=None means something different than always=False
                as_dict["always"] = identity.always
                if identity.on_null is not None:
                    as_dict["on_null"] = identity.on_null
            # attributes common to Identity and Sequence
            attrs = (
                "start",
                "increment",
                "minvalue",
                "maxvalue",
                "nominvalue",
                "nomaxvalue",
                "cycle",
                "cache",
                "order",
            )
            as_dict.update(
                {
                    key: getattr(identity, key, None)
                    for key in attrs
                    if getattr(identity, key, None) is not None
                }
            )
        return as_dict

    has_identity = True

if sqla_2:
    from sqlalchemy.sql.base import _NoneName
else:
    from sqlalchemy.util import symbol as _NoneName  # type: ignore[assignment]


_ConstraintName = Union[None, str, _NoneName]

_ConstraintNameDefined = Union[str, _NoneName]


def constraint_name_defined(
    name: _ConstraintName,
) -> TypeGuard[_ConstraintNameDefined]:
    return name is _NONE_NAME or isinstance(name, (str, _NoneName))


def constraint_name_string(
    name: _ConstraintName,
) -> TypeGuard[str]:
    return isinstance(name, str)


def constraint_name_or_none(
    name: _ConstraintName,
) -> Optional[str]:
    return name if constraint_name_string(name) else None


AUTOINCREMENT_DEFAULT = "auto"


@contextlib.contextmanager
def _ensure_scope_for_ddl(
    connection: Optional[Connection],
) -> Iterator[None]:
    try:
        in_transaction = connection.in_transaction  # type: ignore[union-attr]
    except AttributeError:
        # catch for MockConnection, None
        in_transaction = None
        pass

    # yield outside the catch
    if in_transaction is None:
        yield
    else:
        if not in_transaction():
            assert connection is not None
            with connection.begin():
                yield
        else:
            yield


def url_render_as_string(url, hide_password=True):
    if sqla_14:
        return url.render_as_string(hide_password=hide_password)
    else:
        return url.__to_string__(hide_password=hide_password)


def _safe_begin_connection_transaction(
    connection: Connection,
) -> Transaction:
    transaction = _get_connection_transaction(connection)
    if transaction:
        return transaction
    else:
        return connection.begin()


def _safe_commit_connection_transaction(
    connection: Connection,
) -> None:
    transaction = _get_connection_transaction(connection)
    if transaction:
        transaction.commit()


def _safe_rollback_connection_transaction(
    connection: Connection,
) -> None:
    transaction = _get_connection_transaction(connection)
    if transaction:
        transaction.rollback()


def _get_connection_in_transaction(connection: Optional[Connection]) -> bool:
    try:
        in_transaction = connection.in_transaction  # type: ignore
    except AttributeError:
        # catch for MockConnection
        return False
    else:
        return in_transaction()


def _idx_table_bound_expressions(idx: Index) -> Iterable[ColumnElement[Any]]:
    return idx.expressions  # type: ignore


def _copy(schema_item: _CE, **kw) -> _CE:
    if hasattr(schema_item, "_copy"):
        return schema_item._copy(**kw)  # type: ignore[union-attr]
    else:
        return schema_item.copy(**kw)  # type: ignore[union-attr]


def _get_connection_transaction(
    connection: Connection,
) -> Optional[Transaction]:
    if sqla_14:
        return connection.get_transaction()
    else:
        r = connection._root  # type: ignore[attr-defined]
        return r._Connection__transaction


def _create_url(*arg, **kw) -> url.URL:
    if hasattr(url.URL, "create"):
        return url.URL.create(*arg, **kw)
    else:
        return url.URL(*arg, **kw)


def _connectable_has_table(
    connectable: Connection, tablename: str, schemaname: Union[str, None]
) -> bool:
    if sqla_14:
        return inspect(connectable).has_table(tablename, schemaname)
    else:
        return connectable.dialect.has_table(
            connectable, tablename, schemaname
        )


def _exec_on_inspector(inspector, statement, **params):
    if sqla_14:
        with inspector._operation_context() as conn:
            return conn.execute(statement, params)
    else:
        return inspector.bind.execute(statement, params)


def _nullability_might_be_unset(metadata_column):
    if not sqla_14:
        return metadata_column.nullable
    else:
        from sqlalchemy.sql import schema

        return (
            metadata_column._user_defined_nullable is schema.NULL_UNSPECIFIED
        )


def _server_default_is_computed(*server_default) -> bool:
    if not has_computed:
        return False
    else:
        return any(isinstance(sd, Computed) for sd in server_default)


def _server_default_is_identity(*server_default) -> bool:
    if not sqla_14:
        return False
    else:
        return any(isinstance(sd, Identity) for sd in server_default)


def _table_for_constraint(constraint: Constraint) -> Table:
    if isinstance(constraint, ForeignKeyConstraint):
        table = constraint.parent
        assert table is not None
        return table  # type: ignore[return-value]
    else:
        return constraint.table


def _columns_for_constraint(constraint):
    if isinstance(constraint, ForeignKeyConstraint):
        return [fk.parent for fk in constraint.elements]
    elif isinstance(constraint, CheckConstraint):
        return _find_columns(constraint.sqltext)
    else:
        return list(constraint.columns)


def _reflect_table(inspector: Inspector, table: Table) -> None:
    if sqla_14:
        return inspector.reflect_table(table, None)
    else:
        return inspector.reflecttable(  # type: ignore[attr-defined]
            table, None
        )


def _resolve_for_variant(type_, dialect):
    if _type_has_variants(type_):
        base_type, mapping = _get_variant_mapping(type_)
        return mapping.get(dialect.name, base_type)
    else:
        return type_


if hasattr(sqltypes.TypeEngine, "_variant_mapping"):

    def _type_has_variants(type_):
        return bool(type_._variant_mapping)

    def _get_variant_mapping(type_):
        return type_, type_._variant_mapping

else:

    def _type_has_variants(type_):
        return type(type_) is sqltypes.Variant

    def _get_variant_mapping(type_):
        return type_.impl, type_.mapping


def _fk_spec(constraint):
    source_columns = [
        constraint.columns[key].name for key in constraint.column_keys
    ]

    source_table = constraint.parent.name
    source_schema = constraint.parent.schema
    target_schema = constraint.elements[0].column.table.schema
    target_table = constraint.elements[0].column.table.name
    target_columns = [element.column.name for element in constraint.elements]
    ondelete = constraint.ondelete
    onupdate = constraint.onupdate
    deferrable = constraint.deferrable
    initially = constraint.initially
    return (
        source_schema,
        source_table,
        source_columns,
        target_schema,
        target_table,
        target_columns,
        onupdate,
        ondelete,
        deferrable,
        initially,
    )


def _fk_is_self_referential(constraint: ForeignKeyConstraint) -> bool:
    spec = constraint.elements[0]._get_colspec()  # type: ignore[attr-defined]
    tokens = spec.split(".")
    tokens.pop(-1)  # colname
    tablekey = ".".join(tokens)
    assert constraint.parent is not None
    return tablekey == constraint.parent.key


def _is_type_bound(constraint: Constraint) -> bool:
    # this deals with SQLAlchemy #3260, don't copy CHECK constraints
    # that will be generated by the type.
    # new feature added for #3260
    return constraint._type_bound  # type: ignore[attr-defined]


def _find_columns(clause):
    """locate Column objects within the given expression."""

    cols = set()
    traverse(clause, {}, {"column": cols.add})
    return cols


def _remove_column_from_collection(
    collection: ColumnCollection, column: Union[Column[Any], ColumnClause[Any]]
) -> None:
    """remove a column from a ColumnCollection."""

    # workaround for older SQLAlchemy, remove the
    # same object that's present
    assert column.key is not None
    to_remove = collection[column.key]

    # SQLAlchemy 2.0 will use more ReadOnlyColumnCollection
    # (renamed from ImmutableColumnCollection)
    if hasattr(collection, "_immutable") or hasattr(collection, "_readonly"):
        collection._parent.remove(to_remove)
    else:
        collection.remove(to_remove)


def _textual_index_column(
    table: Table, text_: Union[str, TextClause, ColumnElement[Any]]
) -> Union[ColumnElement[Any], Column[Any]]:
    """a workaround for the Index construct's severe lack of flexibility"""
    if isinstance(text_, str):
        c = Column(text_, sqltypes.NULLTYPE)
        table.append_column(c)
        return c
    elif isinstance(text_, TextClause):
        return _textual_index_element(table, text_)
    elif isinstance(text_, _textual_index_element):
        return _textual_index_column(table, text_.text)
    elif isinstance(text_, sql.ColumnElement):
        return _copy_expression(text_, table)
    else:
        raise ValueError("String or text() construct expected")


def _copy_expression(expression: _CE, target_table: Table) -> _CE:
    def replace(col):
        if (
            isinstance(col, Column)
            and col.table is not None
            and col.table is not target_table
        ):
            if col.name in target_table.c:
                return target_table.c[col.name]
            else:
                c = _copy(col)
                target_table.append_column(c)
                return c
        else:
            return None

    return visitors.replacement_traverse(  # type: ignore[call-overload]
        expression, {}, replace
    )


class _textual_index_element(sql.ColumnElement):
    """Wrap around a sqlalchemy text() construct in such a way that
    we appear like a column-oriented SQL expression to an Index
    construct.

    The issue here is that currently the Postgresql dialect, the biggest
    recipient of functional indexes, keys all the index expressions to
    the corresponding column expressions when rendering CREATE INDEX,
    so the Index we create here needs to have a .columns collection that
    is the same length as the .expressions collection.  Ultimately
    SQLAlchemy should support text() expressions in indexes.

    See SQLAlchemy issue 3174.

    """

    __visit_name__ = "_textual_idx_element"

    def __init__(self, table: Table, text: TextClause) -> None:
        self.table = table
        self.text = text
        self.key = text.text
        self.fake_column = schema.Column(self.text.text, sqltypes.NULLTYPE)
        table.append_column(self.fake_column)

    def get_children(self):
        return [self.fake_column]


@compiles(_textual_index_element)
def _render_textual_index_column(
    element: _textual_index_element, compiler: SQLCompiler, **kw
) -> str:
    return compiler.process(element.text, **kw)


class _literal_bindparam(BindParameter):
    pass


@compiles(_literal_bindparam)
def _render_literal_bindparam(
    element: _literal_bindparam, compiler: SQLCompiler, **kw
) -> str:
    return compiler.render_literal_bindparam(element, **kw)


def _column_kwargs(col: Column) -> Mapping:
    if sqla_13:
        return col.kwargs
    else:
        return {}


def _get_constraint_final_name(
    constraint: Union[Index, Constraint], dialect: Optional[Dialect]
) -> Optional[str]:
    if constraint.name is None:
        return None
    assert dialect is not None
    if sqla_14:
        # for SQLAlchemy 1.4 we would like to have the option to expand
        # the use of "deferred" names for constraints as well as to have
        # some flexibility with "None" name and similar; make use of new
        # SQLAlchemy API to return what would be the final compiled form of
        # the name for this dialect.
        return dialect.identifier_preparer.format_constraint(
            constraint, _alembic_quote=False
        )
    else:
        # prior to SQLAlchemy 1.4, work around quoting logic to get at the
        # final compiled name without quotes.
        if hasattr(constraint.name, "quote"):
            # might be quoted_name, might be truncated_name, keep it the
            # same
            quoted_name_cls: type = type(constraint.name)
        else:
            quoted_name_cls = quoted_name

        new_name = quoted_name_cls(str(constraint.name), quote=False)
        constraint = constraint.__class__(name=new_name)

        if isinstance(constraint, schema.Index):
            # name should not be quoted.
            d = dialect.ddl_compiler(dialect, None)  # type: ignore[arg-type]
            return d._prepared_index_name(  # type: ignore[attr-defined]
                constraint
            )
        else:
            # name should not be quoted.
            return dialect.identifier_preparer.format_constraint(constraint)


def _constraint_is_named(
    constraint: Union[Constraint, Index], dialect: Optional[Dialect]
) -> bool:
    if sqla_14:
        if constraint.name is None:
            return False
        assert dialect is not None
        name = dialect.identifier_preparer.format_constraint(
            constraint, _alembic_quote=False
        )
        return name is not None
    else:
        return constraint.name is not None


def _is_mariadb(mysql_dialect: Dialect) -> bool:
    if sqla_14:
        return mysql_dialect.is_mariadb  # type: ignore[attr-defined]
    else:
        return bool(
            mysql_dialect.server_version_info
            and mysql_dialect._is_mariadb  # type: ignore[attr-defined]
        )


def _mariadb_normalized_version_info(mysql_dialect):
    return mysql_dialect._mariadb_normalized_version_info


def _insert_inline(table: Union[TableClause, Table]) -> Insert:
    if sqla_14:
        return table.insert().inline()
    else:
        return table.insert(inline=True)  # type: ignore[call-arg]


if sqla_14:
    from sqlalchemy import create_mock_engine
    from sqlalchemy import select as _select
else:
    from sqlalchemy import create_engine

    def create_mock_engine(url, executor, **kw):  # type: ignore[misc]
        return create_engine(
            "postgresql://", strategy="mock", executor=executor
        )

    def _select(*columns, **kw) -> Select:  # type: ignore[no-redef]
        return sql.select(list(columns), **kw)  # type: ignore[call-overload]


def is_expression_index(index: Index) -> bool:
    for expr in index.expressions:
        if is_expression(expr):
            return True
    return False


def is_expression(expr: Any) -> bool:
    while isinstance(expr, UnaryExpression):
        expr = expr.element
    if not isinstance(expr, ColumnClause) or expr.is_literal:
        return True
    return False



#!/usr/bin/env python3

# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import REMAINDER, ArgumentParser
from logging import getLogger
from os import environ, execl, getcwd
from os.path import abspath, dirname, pathsep
from re import sub
from shutil import which

from pkg_resources import iter_entry_points

from opentelemetry.instrumentation.version import __version__

_logger = getLogger(__name__)


def run() -> None:
    parser = ArgumentParser(
        description="""
        opentelemetry-instrument automatically instruments a Python
        program and its dependencies and then runs the program.
        """,
        epilog="""
        Optional arguments (except for --help and --version) for opentelemetry-instrument
        directly correspond with OpenTelemetry environment variables. The
        corresponding optional argument is formed by removing the OTEL_ or
        OTEL_PYTHON_ prefix from the environment variable and lower casing the
        rest. For example, the optional argument --attribute_value_length_limit
        corresponds with the environment variable
        OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT.

        These optional arguments will override the current value of the
        corresponding environment variable during the execution of the command.
        """,
    )

    argument_otel_environment_variable = {}

    for entry_point in iter_entry_points(
        "opentelemetry_environment_variables"
    ):
        environment_variable_module = entry_point.load()

        for attribute in dir(environment_variable_module):
            if attribute.startswith("OTEL_"):
                argument = sub(r"OTEL_(PYTHON_)?", "", attribute).lower()

                parser.add_argument(
                    f"--{argument}",
                    required=False,
                )
                argument_otel_environment_variable[argument] = attribute

    parser.add_argument(
        "--version",
        help="print version information",
        action="version",
        version="%(prog)s " + __version__,
    )
    parser.add_argument("command", help="Your Python application.")
    parser.add_argument(
        "command_args",
        help="Arguments for your application.",
        nargs=REMAINDER,
    )

    args = parser.parse_args()

    for argument, otel_environment_variable in (
        argument_otel_environment_variable
    ).items():
        value = getattr(args, argument)
        if value is not None:
            environ[otel_environment_variable] = value

    python_path = environ.get("PYTHONPATH")

    if not python_path:
        python_path = []

    else:
        python_path = python_path.split(pathsep)

    cwd_path = getcwd()

    # This is being added to support applications that are being run from their
    # own executable, like Django.
    # FIXME investigate if there is another way to achieve this
    if cwd_path not in python_path:
        python_path.insert(0, cwd_path)

    filedir_path = dirname(abspath(__file__))

    python_path = [path for path in python_path if path != filedir_path]

    python_path.insert(0, filedir_path)

    environ["PYTHONPATH"] = pathsep.join(python_path)

    executable = which(args.command)
    execl(executable, executable, *args.command_args)



import json
import time
import printer_text_formatter
import printer_wrapper
import printerconstants
from qrcode import (
    QRCode,
    constants,
)

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from cStringIO import StringIO
from datetime import datetime
from logging import Logger

import sysactions
from mw_helper import get_modifier_prefixes

from helper import remove_accents
from messagebus import MessageBus
from production.model import (
    Item,
    ProductionOrder,
    InvalidOrderToPrint,
)
from production.util import (
    rollback_printer_default_font,
    REPRINT_CONSTANT,
)
from production.view.printerviewservices import (
    MwOrderStatus,
)

from xml.etree import cElementTree as eTree


small = printerconstants.TXT_FONT_A
medium = printerconstants.TXT_FONT_SELECT_C + printerconstants.TXT_FONT_C
large = printerconstants.TXT_SIZE + chr(printerconstants.TXT_WIDTH[2] + printerconstants.TXT_HEIGHT[2])
x_large = printerconstants.TXT_SIZE + chr(printerconstants.TXT_WIDTH[5] + printerconstants.TXT_HEIGHT[5])

font_sizes = {1: small, 2: medium, 3: large, 4: x_large}
X_LARGE_LINE_SIZE = 8


class PickListReportGenerator:
    def __init__(
        self,
        message_bus,                # type: MessageBus
        configuration_json_path,    # type: str
        picklist_name,              # type: str
        logger                      # type: Logger
    ):
        # type: (...) -> None

        sysactions.mbcontext = message_bus.mbcontext
        self.configuration_json_path = configuration_json_path
        self.logger = logger
        self.font_size = "small"
        self.picklist_font_size_items = None
        self.picklist_name = picklist_name

        self.item_level_separator_stack = []
        self.last_item_level = 0
        self.configuration_json = self._get_configuration_json()
        self.line_sizes_by_font_size = self.configuration_json["lineSize"]
        self.modifiers_prefixes = get_modifier_prefixes()

    def _get_configuration_json(self):
        try:
            with open(self.configuration_json_path) as json_file:
                return json.loads(json_file.read())
        except Exception:  # noqa
            self.logger.exception(
                "Error opening picklist config file: %r",
                self.configuration_json_path,
            )
            with open("../src/new-production/static/default_picklist_config.json") as json_file:
                return json.loads(json_file.read())

    def production_pick_list(
        self,
        production_order,   # type: ProductionOrder
        is_reprint,         # type: bool
    ):
        try:
            if self._order_is_voided(production_order=production_order):
                self.logger.warning("ignoring voided order")
                raise InvalidOrderToPrint()

            if self._order_is_kiosk(pod_type=production_order.pod_type):
                self.logger.warning("ignoring kiosk order")
                raise InvalidOrderToPrint()

            report_header = self._get_report_header(
                production_order=production_order,
                picklist_name=self.picklist_name,
            )
            order_comment = self._generate_comment(
                production_order=production_order,
                key='COMMENT',
            )
            tab_add_info_comment = self._generate_comment(
                production_order=production_order,
                key='TAB_ADD_INFO',
                top_separator=True,
            )

            sale_line_by_seat_dict = self._get_sale_lines_by_seat(production_order)
            report_sale_type = self._generate_report_sale_type(production_order)
            report_footer = self._get_report_footer(
                production_order=production_order,
            )

            return self.generate_report(
                sale_line_by_seat_dict=sale_line_by_seat_dict,
                order_comment=order_comment,
                report_header=report_header,
                report_footer=report_footer,
                report_sale_type=report_sale_type,
                tab_add_info_comment=tab_add_info_comment,
                is_reprint=is_reprint,
            )
        except (Exception,):
            self.logger.exception("Error building PickList for Order {}".format(production_order.order_id))

    def generate_report(
        self,
        sale_line_by_seat_dict,     # type: dict
        order_comment,              # type: str
        report_header,              # type: str
        report_footer,              # type: str
        report_sale_type,           # type: str
        tab_add_info_comment,       # type: str
        is_reprint,                 # type: bool
    ):
        # type: (...) -> str

        report = StringIO()
        seat_separator = self.configuration_json["body"]["separatorBySeat"]
        first_seat = True
        sale_line_counter = 0
        if "cutMainItems" in self.configuration_json and self.configuration_json["cutMainItems"]:
            for seat_text in sorted(sale_line_by_seat_dict):
                list_sale_line_by_level = sale_line_by_seat_dict[seat_text]

                self._write_header(order_comment, report, report_header)
                items = [item for item in list_sale_line_by_level]

                report.write(self.get_report_body(
                    seat_text,
                    items,
                ))

                report.write(tab_add_info_comment)
                report.write(report_sale_type)
                report.write(rollback_printer_default_font())
                report.write("\n\n")

                report.write(report_footer)
                report.write(rollback_printer_default_font())

                sale_line_counter += 1
                if sale_line_counter < len(sale_line_by_seat_dict):
                    if is_reprint:
                        report.write("\n\n{}".format(REPRINT_CONSTANT))

                    report.write("\n\n\n\n\n")
                    report.write(printerconstants.PAPER_FULL_CUT)
                else:
                    if is_reprint:
                        report.write("\n\n{}\n".format(REPRINT_CONSTANT))
        else:
            self._write_header(order_comment, report, report_header)

            for seat_text in sorted(sale_line_by_seat_dict):
                if first_seat:
                    first_seat = False
                else:
                    if seat_separator:
                        report.write(seat_separator * printerconstants.COLS + "\n")
                list_sale_line_by_level = sale_line_by_seat_dict[seat_text]
                report.write(
                    self.get_report_body(
                        sale_lines=list_sale_line_by_level,
                        seat_text=seat_text
                    )
                )
            report.write(tab_add_info_comment)
            report.write(report_sale_type)

            report.write(report_footer)
            if is_reprint:
                report.write("\n\n{}\n".format(REPRINT_CONSTANT))
        return report.getvalue()

    @staticmethod
    def _write_header(order_comment, report, report_header):
        report.write(printerconstants.TXT_ALIGN_LT)
        report.write(printerconstants.TXT_BOLD_ON)
        report.write(report_header)
        report.write(printerconstants.TXT_BOLD_OFF)
        report.write(order_comment)

    def _get_report_header(
        self,
        production_order,   # type: ProductionOrder
        picklist_name       # type: str
    ):
        # type: (...) -> str

        header_pod_type = production_order.pod_type
        header_sale_type = production_order.sale_type
        totem_print = header_pod_type == "TT" or header_sale_type == "TOTEM"
        drive_print = header_pod_type == "DT" or header_sale_type == "DRIVE_THRU"
        delivery_print = header_pod_type == "DL" or header_sale_type == "DELIVERY"
        sale_type_text = self._get_sale_type(
            delivery_print=delivery_print,
            drive_print=drive_print,
            production_order=production_order,
            totem_print=totem_print
        )

        header_config = self.configuration_json.get("header")
        if not header_config:
            return ""

        header_separator = (header_config.get("separator", "") * printerconstants.COLS)
        produced_datetime = ""
        timestamp = self._get_first_printed_time(production_order)
        if timestamp is not None:
            produced_datetime = time.strftime(printerconstants.DATE_TIME_FMT, timestamp.timetuple())

        pos_id = int(production_order.originator[-2:])
        user_info = "{} / {}".format(production_order.operator_id, production_order.operator_name)

        order_id = int(production_order.order_id.split("-")[0]) % 10000
        parent_table_id = None
        table_name = self._get_property_value(
            production_order=production_order,
            key="TAB_ID"
        )
        order_parent_descr = None
        order_parent_descr_id = None
        if table_name is not None:
            order_descr = "Comanda"
            order_descr_id = printer_text_formatter.get_text_ellipsis(
                text_id=order_descr,
                text=table_name,
                padding_space=3
            )
            parent_table_id = self._get_property_value(
                production_order=production_order,
                key="PARENT_ID"
            )
            if parent_table_id is not None:
                order_parent_descr = "Mesa"
                order_parent_descr_id = printer_text_formatter.get_text_ellipsis(
                    text_id=order_parent_descr,
                    text=parent_table_id,
                    padding_space=3
                )
        else:
            table_name = self._get_property_value(
                production_order=production_order,
                key="TABLE_NAME"
            )
            if table_name is not None:
                order_descr = "Mesa"
                order_descr_id = printer_text_formatter.get_text_ellipsis(
                    text_id=order_descr,
                    text=table_name,
                    padding_space=3
                )
            else:
                order_descr = "Pedido"
                order_descr_id = "{}".format(order_id)

        created_at = time.strptime(production_order.created_at, "%Y-%m-%dT%H:%M:%S.%f")
        created_at = time.strftime(printerconstants.DATE_TIME_FMT, created_at)

        header = self._get_free_label_text(header_config)

        header += font_sizes[header_config["fontSize"]]
        line_size = self.line_sizes_by_font_size.get(str(header_config["fontSize"]), 42)
        if header_config["blackBackground"]:
            header += printerconstants.TXT_INVERT_ON

        if parent_table_id:
            header += printer_text_formatter.center_text(
                text="{}: {}".format(order_parent_descr, order_parent_descr_id),
                line_size=line_size
            )
            header += "\n"

        is_customer_name_used = False
        customer_name = self._get_property_value(
            production_order=production_order,
            key="CUSTOMER_NAME",
        )

        is_pager_number_used = False
        pager_number = self._get_property_value(
            production_order=production_order,
            key="PAGER_NUMBER",
        )

        order_partner_id = self._get_property_value(
            production_order=production_order,
            key="SHORT_REFERENCE",
        )

        header += self._get_reference_number(
            customer_name=customer_name,
            header_config=header_config,
            order_descr=order_descr,
            order_descr_id=order_descr_id,
            pager_number=pager_number,
            order_partner_id=order_partner_id,
        )

        header += self._get_origin_order(
            header_config=header_config,
            production_order=production_order,
        )

        header += self._get_delivery_brand(
            header_config=header_config,
            production_order=production_order,
        )

        header += "\n{}".format(header_separator)

        header += "\n{}".format(self._get_formatted_kds_title(
            picklist_name,
            sale_type_text,
        ))

        header += font_sizes[header_config["fontSize"]]
        if header_config["enabledLabels"]["orderNumber"]:
            header += "\nCodigo........: %s" % order_id

        if header_config["enabledLabels"]["posNumber"]:
            header += "\nPOS...........: %s" % str(pos_id).zfill(2)

        if header_config.get("enabledLabels", {}).get("partnerId", None):
            header += self._print_delivery_partner(
                delivery_print=delivery_print,
                production_order=production_order,
            )

        if pager_number and not is_pager_number_used:
            header += "\nPager.........: %s" % pager_number

        if customer_name and not is_customer_name_used:
            header += "\nCliente.......: %s" % customer_name

        current_datetime = time.strftime(printerconstants.DATE_TIME_FMT, time.localtime())
        if header_config["enabledLabels"]["orderDatetime"]:
            if produced_datetime:
                header += "\nHora Pedido...: %s" % created_at
            else:
                header += "\nImpresso em...: {}".format(current_datetime)

        if header_config["enabledLabels"]["productionDatetime"] and produced_datetime:
            header += "\nHora Produzido: %s" % produced_datetime

        if header_config["enabledLabels"]["orderType"]:
            pos_id = int(production_order.originator[-2:])
            model = sysactions.get_model(posid=pos_id)
            order_type_translation = sysactions.translate_message(
                model=model,
                message_key=production_order.order_type,
            )
            header += "\nTipo de Pedido: %s" % order_type_translation or production_order.order_type

        if header_config["enabledLabels"]["operator"]:
            header += "\nOperador......: %s" % user_info

        header += "\n{}".format(header_separator)

        if header_config.get("printAddress"):
            header += self._print_address(
                delivery_print=delivery_print,
                production_order=production_order,
                separator=header_separator
            )

        header += rollback_printer_default_font()

        return header

    def _get_origin_order(
        self,
        header_config,      # type: dict
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        has_order_origin = self.has_property_in_config(config=header_config, property_name="orderOrigin")
        if not has_order_origin:
            return ""

        model = sysactions.get_model(posid=0)
        order_reference = sysactions.translate_message(
            model=model,
            message_key=production_order.sale_type,
        )

        if production_order.sale_type == "EAT_IN":
            company_id = sysactions.get_storewide_config(
                parameter="Store.CompanyId",
                defval="",
            )

            store_id = sysactions.get_storewide_config(
                parameter="Store.Id",
                defval="",
            )
            order_reference = "{}: {} {}".format(order_reference, company_id, store_id)
        elif production_order.sale_type == "DELIVERY":
            order_partner = self._get_property_value(
                production_order=production_order,
                key="PARTNER",
            )
            order_reference = "{} {}".format(order_reference, order_partner)

        order_reference = self._get_formatted_text(
            text=order_reference,
            style_config=header_config["orderOrigin"],
        )

        header = order_reference + "\n"
        header += rollback_printer_default_font()
        return header

    def _get_delivery_brand(
        self,               # type: PickListReportGenerator
        header_config,      # type: dict
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        has_delivery_brand = self.has_property_in_config(config=header_config, property_name="useDeliveryBrand")
        if not has_delivery_brand:
            return ""

        delivery_brand = self._get_property_value(
            production_order=production_order,
            key="BRAND",
        )

        if delivery_brand in [None, '']:
            return ""

        delivery_brand = self._get_formatted_text(
            text=delivery_brand,
            style_config=header_config.get("deliveryBrand", ""),
        )

        header = delivery_brand + "\n"
        header += rollback_printer_default_font()
        return header

    def _get_reference_number(
        self,               # type: PickListReportGenerator
        customer_name,      # type: str
        header_config,      # type: dict
        order_descr,        # type: str
        order_descr_id,     # type: str
        pager_number,       # type: str
        order_partner_id,   # type: str
    ):
        # type: (...) -> str

        enabled_labels = header_config.get("enabledLabels", {})
        has_reference_number = self.has_property_in_config(config=enabled_labels, property_name="referenceNumber")
        if not has_reference_number:
            return ""

        reference_number_text = "%s: %s" % (order_descr, order_descr_id)

        if header_config.get("usePartnerReferenceNumber") and order_partner_id not in [None, '']:
            reference_number_text = order_partner_id
        elif header_config.get("usePagerAsReferenceNumber") and pager_number:
            reference_number_text = "%s: %s" % ("Pager", pager_number)
        elif header_config.get("useCustomerNameAsReferenceNumber") and customer_name:
            reference_number_text = customer_name

        header = self._get_formatted_reference_number(text=reference_number_text) + "\n"
        header += rollback_printer_default_font()
        return header

    def _get_report_footer(
        self,
        production_order,    # type: ProductionOrder
    ):
        # type: (...) -> str

        footer_config = self.configuration_json.get("footer")
        if not footer_config:
            return ""

        footer = "\n\n" + self._get_free_label_text(json_config=footer_config)

        show_qr_code = self.has_property_in_config(config=footer_config, property_name="showQrCode")
        if show_qr_code:
            footer += self._get_qr_code(production_order=production_order)

        show_debug_line = self.has_property_in_config(config=footer_config, property_name="showDebugLine")
        if show_debug_line:
            footer += self._get_debug_line(
                production_order=production_order
            )

        return footer

    def _get_debug_line(
        self,
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        table_or_tab_name = self._get_table_tab_name(production_order=production_order)
        debug_line = "{}|{}|{}|{}|{}".format(
            production_order.order_id,
            production_order.pos_id,
            datetime.now().strftime("%d%m%Y%H%M%S"),
            production_order.operator_id,
            table_or_tab_name,
        )
        debug_line_text = printer_text_formatter.center_text(
            text=debug_line,
        )
        debug_line_text += rollback_printer_default_font()
        return debug_line_text

    def _get_free_label_text(self, json_config):
        free_label = json_config.get("freeLabel")
        if not free_label:
            return ""

        footer = ""
        if free_label.get("text", "") != "":
            if free_label["blackBackground"]:
                footer += printerconstants.TXT_INVERT_ON

            if free_label["bold"]:
                footer += printerconstants.TXT_BOLD_ON

            footer += font_sizes[free_label["fontSize"]]
            footer += printer_text_formatter.center_text(
                text=free_label["text"].encode("cp860"),
            ) + "\n"
            footer += rollback_printer_default_font()

        return footer

    def get_report_body(
        self,
        seat_text,  # type: str
        sale_lines  # type: list
    ):
        # type: (...) -> str

        body_config = self.configuration_json["body"]
        body = font_sizes[body_config.get("fontSize", 1)]
        body += "\n"
        body += seat_text

        sale_line_separator_char = body_config.get("saleLineSeparator")
        sale_line_separator = ""
        if sale_line_separator_char and len(sale_lines) > 1:
            sale_line_separator = sale_line_separator_char * printerconstants.COLS + "\n"

        for index, product_sale_line in enumerate(sale_lines):
            if index > 0:
                body += sale_line_separator

            if isinstance(product_sale_line, list):
                for sale_line in product_sale_line:
                    if sale_line is None:
                        self.logger.warning("Saleline is None")
                        continue
                    body += sale_line
            else:
                body += product_sale_line

        body += rollback_printer_default_font()
        return body

    def _generate_comment(
        self,
        production_order,       # type: ProductionOrder
        key,                    # type: str
        top_separator=False     # type: Optional[bool]
    ):
        # type: (...) -> str

        report = StringIO()
        comment = self._get_property_value(
            production_order=production_order,
            key=key,
        )
        if comment:
            if top_separator:
                report.write("\n{}".format(printerconstants.SEPARATOR))

            required_lines = len(comment) / printerconstants.COLS + 1
            report.write("\n")
            for line in range(required_lines):
                begin = line * printerconstants.COLS
                end = (line + 1) * printerconstants.COLS
                report.write("\n%s" % comment[begin:end])

            report.write("\n\n{}".format(printerconstants.SEPARATOR))

        return report.getvalue()

    @staticmethod
    def _get_first_printed_time(
        production_order    # type: ProductionOrder
    ):
        # type: (...) -> Optional[datetime]

        for item in production_order.items:
            for tag_history in item.tag_history:
                if tag_history.tag == "served":
                    return tag_history.date
        return None

    def _get_sale_lines_by_seat(
        self,
        production_order        # type: ProductionOrder
    ):
        # type: (...) -> Dict

        grouped_sale_lines = self._get_grouped_sale_lines_by_seat(production_order=production_order)
        if len(grouped_sale_lines) == 0:
            self.logger.warning("Order {} has no sale lines to print".format(production_order.order_id))
            raise Exception()

        return grouped_sale_lines

    def _get_formatted_sale_line(
        self,
        model,                              # type:
        item,                               # type: Item
        pre_option_level=None,              # type: Optional[int]
        preceding_removed_combo=False,      # type: Optional[bool]
        cols=printerconstants.COLS,         # type: Optional[int]
    ):
        # type: (...) -> Tuple[Optional[str], Optional[int], bool]

        item_type = item.item_type

        level = int(item.level)
        level_text = self._get_level_text(level)

        tabulation = self.configuration_json["body"]["layoutByLevel"][level_text]["tabulation"]
        font_size = font_sizes[self.configuration_json["body"]["layoutByLevel"][level_text]["fontSize"]]
        bold = self.configuration_json["body"]["layoutByLevel"][level_text]["bold"]
        black_background = self.configuration_json["body"]["layoutByLevel"][level_text]["blackBackground"]
        highlight_non_production_items = self.configuration_json["body"].get("highlightNonProductionItems", False)

        if item_type == "COMBO":
            preceding_removed_combo = False

        if level == 0:
            pre_option_level = None

        if item_type in [None, "OPTION", "COUPON"]:
            pre_option_level = level if item_type == "OPTION" else None
            return None, pre_option_level, preceding_removed_combo

        include_item_level_separator = False
        level_separator = self.configuration_json["body"]["layoutByLevel"][level_text]["separator"]

        if level_separator != '':
            if level_separator not in self.item_level_separator_stack:
                self.item_level_separator_stack.append(level_separator)

            include_item_level_separator = True

        qty = float(item.qty or 0)
        multiplied_qty = float(item.multiplied_qty or 0)
        default_qty = float(item.default_qty or 0)

        if item_type in ("INGREDIENT", "CANADD", "PRODUCT") and multiplied_qty > 0:
            qty = multiplied_qty

        if (qty == 0 and default_qty == 0) or preceding_removed_combo:
            if item_type == "COMBO":
                preceding_removed_combo = True

            return None, None, preceding_removed_combo

        qty_description = ""
        if qty < default_qty and item.item_type == 'INGREDIENT':
            qty_description = "{} ".format(sysactions.translate_message(
                model=model,
                message_key=self.modifiers_prefixes['[Without]'].replace("$", ""),
            ))

        fraction_parent = item.fraction_parent
        if qty > default_qty and item.level > 0 and item.item_type == 'INGREDIENT':
            qty_description = "{} ".format(sysactions.translate_message(
                model=model,
                message_key=self.modifiers_prefixes['[Additional]'].replace("$", ""),
            ))

        if qty != 0:
            measure_unit = (item.measure_unit or "UN").upper()
            if fraction_parent:
                numerator = int(round(abs(multiplied_qty - default_qty) * fraction_parent))
                formatted_qty = "{}/{}".format(numerator, fraction_parent)
                qty_description += formatted_qty if numerator != fraction_parent else "1".zfill(2)
            else:
                decimal_places = 3 if measure_unit == "KG" else 0
                quantity_format = "{{:.{}f}}".format(decimal_places)
                qty_description = quantity_format.format(qty)
                qty_description = qty_description.zfill(2) if decimal_places == 0 else qty_description

        indentation = tabulation * ((pre_option_level or level) + 1)
        product_name = item.description
        manual_course = item.manual_course
        if manual_course and level == 0:
            manual_course = "[" + manual_course + "]"

        formatted_sale_line = font_size
        if bold:
            formatted_sale_line += printerconstants.TXT_BOLD_ON
        if black_background:
            formatted_sale_line += printerconstants.TXT_INVERT_ON

        if highlight_non_production_items and not self.item_has_jit_line(item=item):
            formatted_sale_line += printerconstants.TXT_INVERT_ON

        description_length = cols - len(manual_course)
        if len(item.comments) > 0:
            for comment in item.comments.values():
                comment_description = ""
                jumped_comment = ""
                if comment.text:
                    if comment.text.startswith("["):
                        comment_description = self._get_configured_comment_description(
                            model=model,
                            comment=comment.text,
                        )
                    else:
                        description = "* " + comment.text
                        indentation_comment_size = len(indentation) + int(round(qty, 1))
                        jumped_comment = self._split_comment_in_lines(
                            comment=description,
                            indentation_size=indentation_comment_size,
                            cols=cols
                        )

                product_description = str(qty_description) + " " + product_name
                if str(comment_description):
                    product_description = str(comment_description) + " " + product_description

                line_rjust = len(product_description) + len(indentation)

                product_description = product_description.rjust(line_rjust)
                if manual_course:
                    product_description = product_description.ljust(description_length)[:description_length]

                description = remove_accents(text=product_description + manual_course)

                formatted_sale_line += description + "\n"
                if jumped_comment:
                    formatted_sale_line += jumped_comment + "\n"
        else:
            product_description = indentation + qty_description + " " + product_name
            if manual_course:
                product_description = product_description.ljust(description_length)[:description_length]
            formatted_sale_line += remove_accents(text=product_description + manual_course) + "\n"

        if include_item_level_separator:
            formatted_sale_line += rollback_printer_default_font()
            formatted_sale_line += level_separator * cols + "\n"

        if level <= pre_option_level:
            pre_option_level = None
        self.last_item_level = level
        formatted_sale_line = self._clean_rollback_default_font(formatted_sale_line)
        return formatted_sale_line, pre_option_level, preceding_removed_combo

    def _get_configured_comment_description(
        self,       # type: PickListReportGenerator
        model,      # type: eTree.Element
        comment,    # type: str
    ):
        # type: (...) -> str

        comment_description = self.modifiers_prefixes.get(comment, comment)
        if "[" in comment_description or "$" in comment_description:
            comment_description = comment_description.replace("[", "").replace("]", "").replace("$", "").upper()
            comment_description = sysactions.translate_message(
                model=model,
                message_key=str(comment_description),
            )
        return comment_description

    @staticmethod
    def _clean_rollback_default_font(
        content,  # type: str
    ):
        # type: (...) -> str

        sequence = rollback_printer_default_font()
        while True:
            if content.endswith("\n"):
                content = content.rstrip("\n")
            elif content.endswith(sequence):
                content = content.rstrip(sequence)
            else:
                break
        return "{}{}\n".format(content, sequence)

    def _get_level_text(self, level):
        current_level = level
        level_text = "level" + str(current_level)
        while level_text not in self.configuration_json["body"]["layoutByLevel"]:
            current_level = current_level - 1
            level_text = "level" + str(current_level)

        return level_text

    def _get_formatted_kds_title(
        self,            # type: PickListReportGenerator
        pick_list_name,  # type: str
        sale_type_text,  # type: str
    ):
        # type: (...) -> str

        header_config = self.configuration_json.get("header", {})
        kds_title_config = header_config.get("kdsTitle", {})
        text = "{}{}".format(pick_list_name, sale_type_text)

        if not kds_title_config:
            return printer_text_formatter.center_text(text=text)

        kds_title = self._get_formatted_text(
            text=text,
            style_config=kds_title_config,
        )
        return kds_title

    def _get_formatted_reference_number(
        self,
        text,  # type: str
    ):
        # type: (...) -> str

        header_config = self.configuration_json.get("header", {})
        reference_number_config = header_config.get("referenceNumber", {})
        reference_number = self._get_formatted_text(
            text=text,
            style_config=reference_number_config,
        )
        return reference_number

    def _get_formatted_text(
        self,           # type: PickListReportGenerator
        text,           # type: str
        style_config,   # type: Dict
    ):
        # type: (...) -> str

        parts = []
        is_rollback_needed = False

        if style_config:
            line_size = printerconstants.COLS
            if style_config.get("blackBackground"):
                parts.append(printerconstants.TXT_INVERT_ON)

            parts.append(printerconstants.TXT_BOLD_OFF)
            if style_config.get("bold"):
                parts.append(printerconstants.TXT_BOLD_ON)

            font_size = style_config.get("fontSize")
            if font_size:
                line_size = self.line_sizes_by_font_size.get(str(font_size), printerconstants.COLS)
                parts.append(font_sizes[font_size])
            is_rollback_needed = bool(parts)

            centered = style_config.get("centered")
            if centered:
                text = printer_text_formatter.center_text(
                    text=text,
                    line_size=line_size,
                )

                if line_size == X_LARGE_LINE_SIZE:
                    parts.append(printerconstants.TXT_ALIGN_CT)

        parts.append(text)

        if is_rollback_needed:
            parts.append(rollback_printer_default_font())

        return "".join(parts)

    @staticmethod
    def _split_comment_in_lines(
        comment,            # type: str
        indentation_size,   # type: int
        cols                # type: int
    ):
        # type: (...) -> str

        line_rjust = len(comment) + indentation_size
        if line_rjust > cols:
            comment = remove_accents(text=comment)
            remaining_size = cols - indentation_size
            lines = [comment[i: i + remaining_size] for i in range(0, len(comment), remaining_size)]
            comment = "\n".join([(" " * indentation_size) + line for line in lines])
        else:
            comment = remove_accents(text=comment.rjust(line_rjust))

        return comment

    @staticmethod
    def _get_title_sale_lines_seats(
        seat,           # type: str
        table_seats,    # type: str
    ):
        # type: (...) -> str

        seat_title = "Assento {}"

        if seat != "":
            title = seat_title.format(seat)
        else:
            title = "Mesa ({} assentos)".format(table_seats) if table_seats else "Sem Assento"

        return "{}\n".format(title)

    def _generate_report_sale_type(
        self,
        production_order    # type: ProductionOrder
    ):
        # type: (...) -> str

        report = StringIO()
        sale_type = ""
        if production_order.pod_type == "DT":
            sale_type = "DRIVE-THRU"
        else:
            if production_order.sale_type != 'EAT_IN':
                sale_type = production_order.sale_type.replace('_', ' ').upper()
                if sale_type == "APP":
                    pickup_type = self._get_property_value(
                        production_order=production_order,
                        key="PICKUP_TYPE"
                    )

                    if pickup_type.upper() == "TAKE_OUT":
                        sale_type = "VIAGEM"

                    if pickup_type.upper() == "EAT_IN":
                        sale_type = ""
                if "TAKE" in sale_type:
                    sale_type = "VIAGEM"

        if sale_type != "":
            sale_type_text = printer_text_formatter.center_text(
                text=sale_type,
                sep='*',
                line_size=printerconstants.COLS / 2,
            )
            sale_type_text = printerconstants.TXT_4SQUARE + sale_type_text + printerconstants.TXT_NORMAL
            report.write("\n%s" % sale_type_text)

        return report.getvalue()

    @staticmethod
    def _create_qrcode(
        qr_data,        # type: str
    ):
        # type: (...) -> str

        qr = QRCode(
            version=2,
            error_correction=constants.ERROR_CORRECT_H,
            box_size=10,
            border=4,
        )

        qr.add_data(qr_data)
        qr.make(fit=True)

        img = qr.make_image()
        img = img.convert("RGB")
        img = printer_wrapper.image(img)

        return img.strip()

    def _print_delivery_partner(
        self,
        delivery_print,     # type: str
        production_order    # type: ProductionOrder
    ):
        # type: (...) -> str

        text = ""
        order_partner = self._get_property_value(
            production_order=production_order,
            key="PARTNER"
        )
        order_partner_id = self._get_property_value(
            production_order=production_order,
            key="SHORT_REFERENCE"
        )
        if delivery_print and order_partner and order_partner != "" and order_partner_id != "":
            text += """\nParceiro......: {0}\nID Parceiro...: {1}""".format(order_partner.upper(), order_partner_id)

        return text

    def _print_address(
        self,
        delivery_print,     # type: bool
        production_order,   # type: ProductionOrder
        separator,          # type: str
    ):
        # type: (...) -> str
        text = ""
        valid_sale_type = production_order.sale_type.upper() not in ["TAKE_OUT", "EAT_IN"]
        if not delivery_print or not valid_sale_type:
            return text

        try:
            street_name_element = self._get_property_value(
                production_order=production_order,
                key="STREET_NAME",
            )
            street_number_element = self._get_property_value(
                production_order=production_order,
                key="STREET_NUMBER",
            )
            if street_name_element:
                text += """\nEndereco......: {0}, {1}\n""".format(street_name_element, street_number_element)
            default_format = """{}: {}\n"""
            text += default_format.format(
                "Complemento...",
                self._get_property_value(
                    production_order=production_order,
                    key="COMPLEMENT",
                )
            )
            reference = self._get_property_value(
                production_order=production_order,
                key="REFERENCE",
            )
            if reference:
                text += default_format.format(
                    "Referencia....",
                    reference,
                )
            text += default_format.format(
                "Bairro........",
                self._get_property_value(
                    production_order=production_order,
                    key="NEIGHBORHOOD",
                )
            )
            text += default_format.format(
                "CEP...........",
                self._get_property_value(
                    production_order=production_order,
                    key="POSTAL_CODE",
                )
            )
            text += default_format.format(
                "Cidade........",
                self._get_property_value(
                    production_order=production_order,
                    key="CITY",
                )
            )
            state = self._get_property_value(
                production_order=production_order,
                key="STATE",
            )

            if state not in ['', None]:
                text += default_format.format(
                    "Estado........",
                    state,
                )

            if text != "":
                text += separator

        except (Exception,):
            self.logger.exception("Error in get address on Delivery picklist")

        return text

    def _get_picklist_font_size_items(self):
        # type: () -> str

        if self.picklist_font_size_items is None:
            try:
                sizes = {"small": small, "medium": medium, "large": large}
                font_size_items = self.font_size
                self.picklist_font_size_items = sizes[font_size_items]
            except (Exception,):
                self.picklist_font_size_items = small

        return self.picklist_font_size_items

    def _get_grouped_sale_lines_by_seat(
        self,
        production_order    # type: ProductionOrder
    ):
        # type: (...) -> Dict

        grouped_sale_lines = {}
        pre_option_level = None
        preceding_removed_combo = False
        self.generate_sale_line_of_all_items(
            production_order=production_order,
            grouped_sale_lines=grouped_sale_lines,
            pre_option_level=pre_option_level,
            preceding_removed_combo=preceding_removed_combo,
            items=production_order.items
        )
        return grouped_sale_lines

    def generate_sale_line_of_all_items(
        self,                       # type: PickListReportGenerator
        production_order,           # type: ProductionOrder
        grouped_sale_lines,         # type: Dict
        pre_option_level,           # type: Optional[int]
        preceding_removed_combo,    # type: bool
        items                       # type: List
    ):
        # type: (...) -> Dict

        comments_by_sale_line = self._comments_by_sale_line(items=items)
        items = self._create_grouped_comments(
            items=items,
            comments_by_sale_line=comments_by_sale_line,
        )
        items_by_seat_and_sale_line = self._items_by_seat_and_sale_line(
            items=items,
            production_order=production_order,
        )

        for seat, items_by_sale_line in items_by_seat_and_sale_line.items():
            grouped_sale_lines[seat] = []
            for sale_line, items in items_by_sale_line.items():
                if items[0].grouped_comment:
                    grouped_sale_lines[seat].extend(self._get_formatted_line_with_surrounded_grouped_comment(
                        items=items,
                        preceding_removed_combo=preceding_removed_combo,
                        production_order=production_order,
                        pre_option_level=pre_option_level,
                    ))
                else:
                    grouped_sale_lines[seat].append(self._get_formatted_items_to_print(
                        items=items,
                        preceding_removed_combo=preceding_removed_combo,
                        production_order=production_order,
                        pre_option_level=pre_option_level,
                    ))

                if seat is not None:
                    self._flush_pending_item_level_separators(
                        seat_text=seat,
                        grouped_sale_lines=grouped_sale_lines,
                    )

                self.item_level_separator_stack[:] = []
                self.last_item_level = 0

        return grouped_sale_lines

    def _flush_pending_item_level_separators(
        self,
        seat_text,           # type: str
        grouped_sale_lines,  # type: Dict[str, List[List[Optional[str]]]]
    ):
        # type: (...) -> None

        if self.item_level_separator_stack:
            item_sale_lines = grouped_sale_lines[seat_text]
            last_item_sale_lines = self._get_last_non_none_sublist(item_sale_lines)
            if last_item_sale_lines:
                sale_line_buf = last_item_sale_lines[-1] or ""
                self.item_level_separator_stack.reverse()
                for sep in self.item_level_separator_stack:
                    sale_line_buf += sep * printerconstants.COLS + "\n"
                last_item_sale_lines[-1] = sale_line_buf

    def generate_sale_line_of_all_items_sons(
        self,
        model,                      # type: object
        production_order,           # type: ProductionOrder
        item_sale_lines,            # type: List[str]
        pre_option_level,           # type: int
        preceding_removed_combo,    # type: bool
        items                       # type: List[Item]
    ):
        # type: (...) -> None

        for item in items:
            if item.is_ingredient() and item.default_qty == item.qty and not item.comments:
                continue

            self.generate_sale_line_by_item(
                model=model,
                grouped_sale_lines=item_sale_lines,
                item=item,
                pre_option_level=pre_option_level,
                preceding_removed_combo=preceding_removed_combo
            )
            if len(item.items) > 0:
                self.generate_sale_line_of_all_items_sons(
                    model=model,
                    production_order=production_order,
                    item_sale_lines=item_sale_lines,
                    pre_option_level=pre_option_level,
                    preceding_removed_combo=preceding_removed_combo,
                    items=item.items
                )

    def generate_sale_line_by_item(
        self,
        model,                      # type: object
        grouped_sale_lines,         # type: List[str]
        item,                       # type: Item
        pre_option_level,           # type: int
        preceding_removed_combo     # type: bool
    ):
        # type: (...) -> None

        sale_line, pre_option_level, preceding_removed_combo = self._get_formatted_sale_line(
            model=model,
            item=item,
            pre_option_level=pre_option_level,
            preceding_removed_combo=preceding_removed_combo,
        )
        grouped_sale_lines.append(sale_line)

    @staticmethod
    def _order_is_voided(
        production_order    # type: ProductionOrder
    ):
        # type: (...) -> bool

        return MwOrderStatus.VOIDED.name == production_order.state

    @staticmethod
    def _join(
        dic1,   # type: Dict
        dic2    # type: Dict
    ):
        # type: (...) -> Dict

        d = dict(dic1)
        d.update(dic2)
        return d

    def _get_sale_type(
        self,
        delivery_print,     # type: bool
        drive_print,        # type: bool
        production_order,   # type: ProductionOrder
        totem_print         # type: bool
    ):
        # type: (...) -> str

        if totem_print:
            return "(TOTEM)"
        if drive_print:
            return "(DRIVE-THRU)"
        if delivery_print:
            sale_type = self._get_sale_type_by_pickup_type(production_order=production_order)
            return sale_type

        sale_type = self._get_by_sale_type(order=production_order)
        return sale_type

    def _get_sale_type_by_pickup_type(
        self,
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        sale_type = " (DELIVERY)"
        pickup_type = self._get_property_value(
            production_order=production_order,
            key="PICKUP_TYPE"
        )
        if pickup_type == "TAKE_OUT":
            sale_type = " (DELIVERY - RETIRADA VIAGEM)"
        elif pickup_type == "EAT_IN":
            sale_type = " (DELIVERY - RETIRADA LOJA)"

        return sale_type

    @staticmethod
    def _get_by_sale_type(
        order   # type: ProductionOrder
    ):
        # type: (...) -> str

        if order.sale_type == "TAKE_OUT":
            return " (VIAGEM)"

        return " (LOJA)"

    @staticmethod
    def _get_property_value(
        production_order,   # type: ProductionOrder
        key                 # type: str
    ):
        # type: (...) -> Any

        if key in production_order.properties:
            return remove_accents(production_order.properties[key])
        return None

    @staticmethod
    def _get_last_non_none_sublist(
        main_list,  # type: List[Union[List[str], None]]
    ):
        # type: (...) -> Optional[List]

        for sublist in list(reversed(main_list)):
            if sublist is not None:
                return sublist

        return None

    def _get_qr_code(
        self,               # type: PickListReportGenerator
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        qr_code_data = "QuickService|{}".format(
            production_order.order_id,
        )
        if production_order.sale_type == "DELIVERY":
            partner = self._get_property_value(
                production_order=production_order,
                key="PARTNER",
            )
            short_reference = self._get_property_value(
                production_order=production_order,
                key="SHORT_REFERENCE",
            )

            delivery_brand = self._get_property_value(
                production_order=production_order,
                key="BRAND",
            )
            delivery_brand = delivery_brand if delivery_brand is not None else ""

            originator = self._get_property_value(
                production_order=production_order,
                key="ORIGINATOR",
            )
            originator = originator.split('|')[0] if originator is not None else ""

            qr_code_data = "{}|{}|{}|{}|{}".format(
                production_order.order_id,
                partner,
                short_reference,
                delivery_brand,
                originator
            )

        qr_code = printerconstants.TXT_ALIGN_CT
        qr_code += self._create_qrcode(qr_data=qr_code_data)
        qr_code += rollback_printer_default_font()
        return qr_code

    def _get_table_tab_name(
        self,
        production_order,   # type: ProductionOrder
    ):
        # type: (...) -> str

        tab_id = self._get_property_value(
            production_order=production_order,
            key="TAB_ID",
        )
        if tab_id is not None:
            return "C{}".format(tab_id)

        table_or_tab_name = ""
        table_name = self._get_property_value(
            production_order=production_order,
            key="TABLE_NAME",
        )
        if table_name is not None:
            table_or_tab_name = "M{}".format(table_name)

        return table_or_tab_name

    @staticmethod
    def item_has_jit_line(
        item,  # type: Item
    ):
        # type: (...) -> bool

        return item.jit_lines != ""

    @staticmethod
    def has_property_in_config(
        config,         # type: Dict
        property_name,  # type: str
    ):
        # type: (...) -> bool

        return config.get(property_name, False)

    @staticmethod
    def _comments_by_sale_line(
        items,  # type: List
    ):
        # type: (...) -> Dict

        comments_by_sale_line = {}
        for item in items:
            line_number = item.line_number
            if line_number not in comments_by_sale_line:
                comments_by_sale_line.update({
                    line_number: {
                        "items_qty": 1,
                    }
                })
            else:
                comments_by_sale_line[line_number]["items_qty"] += 1

            for comment in item.comments.values():
                if comment.text not in comments_by_sale_line[line_number]:
                    comments_by_sale_line[line_number].update({
                        comment.text: {
                            "comment_qty": 1,
                        }
                    })
                else:
                    comments_by_sale_line[line_number][comment.text]["comment_qty"] += 1

        return comments_by_sale_line

    @staticmethod
    def _create_grouped_comments(
        items,                  # type: List
        comments_by_sale_line,  # type: Dict
    ):
        # type: (...) -> List

        for item in items:
            line_number = item.line_number
            for comment_id, comment in item.comments.items():
                total_items_in_line_number = comments_by_sale_line[line_number]["items_qty"]
                if total_items_in_line_number == 1:
                    continue

                if comments_by_sale_line[line_number][comment.text]["comment_qty"] == total_items_in_line_number:
                    del item.comments[comment_id]
                    item.grouped_comment = remove_accents(comment.text)

        return items

    def get_seat_key(
        self,              # type: PickListReportGenerator
        item,              # type: Item
        production_order,  # type: ProductionOrder
    ):
        seat_key = ""
        should_separate_by_seat = self.configuration_json["body"]["separateBySeat"]
        if should_separate_by_seat:
            table_id = self._get_property_value(
                production_order=production_order,
                key="TABLE_ID",
            )
            if table_id is not None:
                table_seats = self._get_property_value(
                    production_order=production_order,
                    key="TABLE_SEATS",
                )
                seat_key = self._get_title_sale_lines_seats(
                    seat=item.seat,
                    table_seats=table_seats,
                )
        return seat_key

    def _items_by_seat_and_sale_line(
        self,              # type: PickListReportGenerator
        items,             # type: List
        production_order,  # type: ProductionOrder
    ):
        # type: (...) -> Dict

        items_by_seat_and_sale_line = {}
        for item in items:
            seat_key = self.get_seat_key(
                item=item,
                production_order=production_order,
            )
            if seat_key not in items_by_seat_and_sale_line:
                items_by_seat_and_sale_line[seat_key] = {}

            sale_line = item.line_number
            if sale_line not in items_by_seat_and_sale_line[seat_key]:
                items_by_seat_and_sale_line[seat_key][sale_line] = [item]
            else:
                items_by_seat_and_sale_line[seat_key][sale_line].append(item)

        return items_by_seat_and_sale_line

    def _get_formatted_line_with_surrounded_grouped_comment(
        self,                       # type: PickListReportGenerator
        items,                      # type: List
        preceding_removed_combo,    # type: bool
        production_order,           # type: ProductionOrder
        pre_option_level,           # type: Optional[int]
    ):
        # type: (...) -> List

        formatted_lines_to_print = []
        grouped_comment_separator = "{}{}".format("=" * printerconstants.COLS, "\n")
        formatted_lines_to_print.append([grouped_comment_separator])

        grouped_comment_text = "{}{}".format(items[0].grouped_comment, "\n")
        formatted_lines_to_print.append([grouped_comment_text])

        formatted_lines_to_print.append(self._get_formatted_items_to_print(
            items=items,
            preceding_removed_combo=preceding_removed_combo,
            production_order=production_order,
            pre_option_level=pre_option_level,
        ))

        formatted_lines_to_print.append([grouped_comment_separator])

        return formatted_lines_to_print

    def _get_formatted_items_to_print(
        self,                       # type: PickListReportGenerator
        items,                      # type: List[Item]
        preceding_removed_combo,    # type: bool
        production_order,           # type: ProductionOrder
        pre_option_level,           # type: Optional[int]
    ):
        # type: (...) -> List

        model = sysactions.get_model(0)

        item_sale_lines = []
        for item in items:
            self.generate_sale_line_by_item(
                model=model,
                grouped_sale_lines=item_sale_lines,
                item=item,
                pre_option_level=pre_option_level,
                preceding_removed_combo=preceding_removed_combo,
            )
            if len(item.items) > 0:
                self.generate_sale_line_of_all_items_sons(
                    model=model,
                    production_order=production_order,
                    item_sale_lines=item_sale_lines,
                    pre_option_level=pre_option_level,
                    preceding_removed_combo=preceding_removed_combo,
                    items=item.items,
                )

        return item_sale_lines

    @staticmethod
    def _order_is_kiosk(
        pod_type,   # type: str
    ):
        # type: (...) -> bool

        return pod_type == "KK"



from application.domain.ui import MessageBox
from sysactions import show_messagebox


class MsgBusMessageBox(MessageBox):

    def show(self, pos_id, message, title='$INFORMATION', icon='info', buttons='$OK', timeout=60000, asynch=False, linefeed='\\', focus=True):
        show_messagebox(
            pos_id,
            message,
            title,
            icon,
            buttons,
            timeout,
            asynch,
            linefeed,
            focus
        )
